import { stateMachine } from "./stateMachine/stateMachine";
import { loadBoardsFromDrupal } from "./drupal";
import { eventBus } from "./eventBus";
import { projectDetailCache } from "./cache";
import { isIdling, lastSyncWithServer, isWebSocketConnected, isDirty } from "../stores/appState";
import { get } from "svelte/store";
import { boards } from "../stores/boardData";
import { toast } from "svelte-french-toast";
import { tick } from "svelte/internal";
import {saveBoardsToDrupal} from "./drupal";

let isSyncing = false;
let lastSyncAttempt = 0;
const MIN_SYNC_INTERVAL = 2000; // 2 seconds minimum between syncs

/**
 * Handles sending board data to the server and triggers a sync after completion.
 * This function manages the save operation and ensures data consistency by
 * fetching fresh data after a save attempt, regardless of success or failure.
 * 
 * @async
 * @throws {Error} When the save operation fails
 */
export async function handleSendDataToServer() {
  if (isSyncing) return;
  isSyncing = true;
  await tick();
  try {
    const boardData = get(boards).map((b) => b.serializeToData());
    console.log('boardData', boardData);
    await toast.promise(saveBoardsToDrupal(boardData), {  
      loading: 'Saving boards to server...',
      success: 'Boards saved to server',
      error: 'Error saving boards to server'
    });
    lastSyncWithServer.set(Date.now());
  } catch (error) {
    projectDetailCache.clear();
    throw error;
  } finally {
    isSyncing = false;
    isDirty.set(true); // Mark as dirty after save to ensure we get latest data
    await checkAndTriggerServerFetchIfNeeded();
  }
}

/**
 * Checks if data needs to be fetched from the server and triggers the fetch if needed.
 */
export async function checkAndTriggerServerFetchIfNeeded() {
  console.log(`[Sync] Starting fetch check. IsSyncing: ${isSyncing}, IsDirty: ${get(isDirty)}`);
  
  if (isSyncing) {
    console.log('[Sync] Already syncing, skipping');
    return;
  }

  const now = Date.now();
  const back_then = get(lastSyncWithServer);
  const isDataDirty = get(isDirty);

  // Skip polling if WebSocket is connected and data is not dirty
  if (get(isWebSocketConnected) && !isDataDirty) {
    console.log('[Sync] WebSocket connected and data not dirty, skipping');
    return;
  }

  if (!isDataDirty && !(stateMachine.activeState == 'viewCurrentBoard' || stateMachine.activeState == 'loading')) {
    console.log(`[Sync] Invalid state: ${stateMachine.activeState}, skipping`);
    return;
  }

  // Check sync interval only if data is not dirty
  if (!isDataDirty) {
    console.log(`[Sync] Time check - Now: ${now}, Last: ${back_then}`);
    const shouldSync = !back_then || (back_then + MIN_SYNC_INTERVAL < now);
    if (!shouldSync) {
      console.log('[Sync] Within sync interval and not dirty, skipping');
      return;
    }
  }

  try {
    console.log('[Sync] Starting data fetch');
    isSyncing = true;
    lastSyncAttempt = now;
    const projectData = await loadBoardsFromDrupal();
    console.log('[Sync] Data received, dispatching to event bus');
    eventBus.dispatch('serverBoardDataRecieve', projectData);
    console.log('[Sync] Updating last sync time');
    lastSyncWithServer.set(now);
    isDirty.set(false); // Clear dirty flag after successful sync
    await tick();
  } catch (error) {
    console.error('[Sync] Error during sync:', error);
    projectDetailCache.clear();
    throw error;
  } finally {
    console.log('[Sync] Completed sync');
    isSyncing = false;
  }
} 