<?php

declare(strict_types=1);

namespace Drupal\pm_app\Service;

use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\notification_server\DTO\ChannelDTO;
use Drupal\notification_server\Service\NotificationServerClientInterface;

/**
 * The Notification Service.
 *
 * Handles all notification-related functionality including:
 * - Generation of client IDs
 * - Generation of channel names
 * - Notification management
 *
 * This is a non-shared service.
 *
 * @see https://symfony.com/doc/current/service_container/shared.html
 */
final class PmAppNotificationService {
  public function __construct(
    private readonly LoggerChannelInterface $logger,
    private readonly ConfigFactoryInterface $configFactory,
    private readonly NotificationServerClientInterface $notificationServer
  ) {}

  /**
   * Get a unique client ID for notifications.
   *
   * @return string
   *   A unique client ID.
   */
  public function getClientId(): string {
    $client_id = 'pm_app:' . session_id();
    $this->notificationServer->generateClientId($client_id);
    return $client_id;
  }

  /**
   * Get a channel name for a specific entity.
   *
   * @param string $entityType
   *   The entity type (e.g., 'pm_project', 'pm_board').
   * @param int|string $entityId
   *   The entity ID.
   * @param string|null $subChannel
   *   Optional sub-channel identifier.
   *
   * @return string
   *   The generated channel name.
   */
  public function getChannel(string $entityType, int|string $entityId, ?string $subChannel = null): string {
    $channelParts = [
      'pm_app',
      $entityType,
      $entityId,
    ];

    if ($subChannel !== null) {
      $channelParts[] = $subChannel;
    }
    $channel = implode(':', $channelParts);
    $ch = new ChannelDTO($channel);
    $this->notificationServer->createChannel($ch);
    return $channel;
  }

  /**
   * Publish a new notification.
   *
   * @param string $channel
   *   The channel to publish the notification to.
   * @param string $message
   *   The notification message content.
   *
   * @return array|null
   *   The notification data if successful, NULL otherwise.
   */
  public function publishNotification(string $channel, string $message): void {
    $this->notificationServer->publishNotification($channel, $message);
  }


  public function grantChannelAccess(string $channel, string $clientId): void {
    $this->notificationServer->grantChannelAccess($channel, $clientId);
  }

  public function revokeChannelAccess(string $channel, string $clientId): void {
    $this->notificationServer->revokeChannelAccess($channel, $clientId);
  }

  public function getWebSocketEndpoint(): string {
    return $this->notificationServer->getWebSocketEndpoint();
  }

}
